/**
 *  txt.ivp
 */
#include "stdafx.h"
#include "../ivr.h"

static LPCWSTR Patterns[] = { L"*.txt", NULL };

class TextRenderer : public IPicture
{
	enum
	{
		FONT_SCALE = 120,
		BACK_COLOR = 0xFFFFFF,
		HALF_COLOR = 0x808080,
		FORE_COLOR = 0x000000,
	};
	inline static BOOL GetUIFont(LOGFONT &lf)
	{
		NONCLIENTMETRICS ncm;
		::ZeroMemory(&ncm, sizeof ncm);
		if (::SystemParametersInfo(SPI_GETNONCLIENTMETRICS, ncm.cbSize = sizeof ncm, &ncm, FALSE)) {
			::MoveMemory(&lf, &ncm.lfMessageFont, sizeof lf);
			return TRUE;
		}
		if (HGDIOBJ hFont = ::GetStockObject(DEFAULT_GUI_FONT))
			return ::GetObject(hFont, sizeof lf, &lf);
		return FALSE;
	}
public:
#pragma region IUnknown
	IFACEMETHODIMP QueryInterface(REFIID riid, void **ppv)
	{
		if (::InlineIsEqualGUID(riid, IID_IPicture) ||
			::InlineIsEqualGUID(riid, IID_IUnknown)) {
			AddRef();
			*ppv = static_cast<IPicture *>(this);
			return S_OK;
		}
		return E_NOINTERFACE;
	}
	IFACEMETHODIMP_(ULONG) AddRef()
	{
		return ++RefCount;
	}
	IFACEMETHODIMP_(ULONG) Release()
	{
		if (--RefCount == 0)
			delete this;
		return RefCount;
	}
#pragma endregion
	
#pragma region IPicture
	IFACEMETHODIMP get_Handle(OLE_HANDLE *) { return E_NOTIMPL; }
	IFACEMETHODIMP get_hPal(OLE_HANDLE *) { return E_NOTIMPL; }
	IFACEMETHODIMP get_Type(SHORT *) { return E_NOTIMPL; }
	IFACEMETHODIMP get_Width(OLE_XSIZE_HIMETRIC *) { return E_NOTIMPL; }
	IFACEMETHODIMP get_Height(OLE_YSIZE_HIMETRIC *) { return E_NOTIMPL; }
	IFACEMETHODIMP Render(HDC hDC, LONG x, LONG y, LONG cx, LONG cy,
		OLE_XPOS_HIMETRIC, OLE_YPOS_HIMETRIC, OLE_XSIZE_HIMETRIC, OLE_YSIZE_HIMETRIC, LPCRECT)
	{
		::SaveDC(hDC);
		::SetBkMode(hDC, TRANSPARENT);
		::SetBkColor(hDC, BACK_COLOR);
		if (Font)
			::SelectObject(hDC, Font);
		RECT rc = { x, y, x + cx, y + cy };
		DRAWTEXTPARAMS dtp = { sizeof dtp, 4, 0, 0, 0 };
		UINT format = DT_NOPREFIX | DT_EXPANDTABS | DT_TABSTOP;
		++rc.left, ++rc.top;
		::SetTextColor(hDC, HALF_COLOR), ::DrawTextExW(hDC, Text, ::SysStringLen(Text), &rc, format, &dtp);
		--rc.left, --rc.top;
		::SetTextColor(hDC, FORE_COLOR), ::DrawTextExW(hDC, Text, ::SysStringLen(Text), &rc, format, &dtp);
		::RestoreDC(hDC, -1);
		return S_OK;
	}
	IFACEMETHODIMP set_hPal(OLE_HANDLE) { return E_NOTIMPL; }
	IFACEMETHODIMP get_CurDC(HDC *) { return E_NOTIMPL; }
	IFACEMETHODIMP SelectPicture(HDC, HDC *, OLE_HANDLE *) { return E_NOTIMPL; }
	IFACEMETHODIMP get_KeepOriginalFormat(BOOL *) { return E_NOTIMPL; }
	IFACEMETHODIMP put_KeepOriginalFormat(BOOL) { return E_NOTIMPL; }
	IFACEMETHODIMP PictureChanged() { return E_NOTIMPL; }
	IFACEMETHODIMP SaveAsFile(LPSTREAM, BOOL, LONG *) { return E_NOTIMPL; }
	IFACEMETHODIMP get_Attributes(DWORD *) { return E_NOTIMPL; }
#pragma endregion
	
private:
	TextRenderer(BSTR strText)
		: RefCount(1), Text(strText), Font(NULL)
	{
		LOGFONT lf;
		if (GetUIFont(lf)) {
			lf.lfHeight = lf.lfHeight * FONT_SCALE / 100;
			lf.lfWeight = FW_SEMIBOLD;
			Font = ::CreateFontIndirect(&lf);
		}
	}
	~TextRenderer()
	{
		::DeleteObject(Font);
		::SysFreeString(Text);
	}
	ULONG RefCount;
	BSTR  Text;
	HFONT Font;
	
public:
	inline static HRESULT Create(UINT CodePage, LPCSTR lpSource, ULONG cbSource, IPicture **ppRenderer)
	{
		int cchText = ::MultiByteToWideChar(CodePage, MB_ERR_INVALID_CHARS, lpSource, cbSource, NULL, 0);
		if (cchText <= 0)
			return S_FALSE;
		BSTR strText = ::SysAllocStringLen(NULL, cchText);
		if (!strText)
			return E_OUTOFMEMORY;
		::MultiByteToWideChar(CodePage, MB_ERR_INVALID_CHARS, lpSource, cbSource, strText, cchText);
		*ppRenderer = new TextRenderer(strText);
		if (!*ppRenderer) {
			::SysFreeString(strText);
			return E_OUTOFMEMORY;
		}
		return S_OK;
	}
};

static HRESULT GetPicture(IStream *Stream, IPicture **ppRenderer)
{
	static const ULONG MAX_BUFFER = 4096;
	LPSTR lpSource = new CHAR[MAX_BUFFER];
	if (!lpSource)
		return E_OUTOFMEMORY;
	ULONG cbSource;
	HRESULT hr = Stream->Read(lpSource, MAX_BUFFER, &cbSource);
	if (FAILED(hr)) {
		delete[] lpSource;
		return hr;
	}
	if (TextRenderer::Create(CP_THREAD_ACP, lpSource, cbSource, ppRenderer) == S_OK) {
		delete[] lpSource;
		return S_OK;
	}
	if (TextRenderer::Create(CP_UTF8, lpSource, cbSource, ppRenderer) == S_OK) {
		delete[] lpSource;
		return S_OK;
	}
	delete[] lpSource;
	return E_FAIL;
}

IVR_API BOOL WINAPI GetIVR1(IVR1 *IVR)
{
	IVR->Patterns = Patterns;
	IVR->GetPicture = GetPicture;
	return TRUE;
}
